==> geturl-1.2.tar.gz: geturl-1.2/html.c <==
/* html.c
   Module for parsing HTML buffers */

#include "config.h"
#include <ctype.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include "geturl.h"
#include "html.h"

/* Match a string against a null-terminated list of identifiers. */
int
idmatch(tag_t *tags, char *tag, char *mod)
{
   int i;

   if (!tag || !mod)
      return 0;
   
   for (i = 0; tags[i].tag; i++)
      if (!strcasecmp(tags[i].tag, tag) && !strcasecmp(tags[i].mod, mod))
	 return 1;
   return 0;
}

/* The following function parses an HTML buffer buf, of size
   bufsize searching for tags. When a tag is encountered, the
   routine will extract its components. If "src", "href" or
   "background" are found, their address is returned, and so
   is the length of the string. Returns NULL if no URL
   is found.

   You must also provide an id for the specified buffer, so
   that the function can associate it with a parser state.
   Thus several buffers can be parsed, and all the switching
   between them will be accurate, provided each of them has
   a unique ID (e.g. its filename).

   Latest: add no-quote support (e.g. href=http://something/)

   Note: possible bugs at the end of buffers, or with small
   buffers */

char *
htmlfindurl(char *buf, int bufsize, int *size, char *id)
{
   char *ptr, **p;
   state_t *s;
   /* NULL-terminated list of tags and modifiers someone would want to
      follow -- feel free to edit to suit your needs: */
   static tag_t allow_tag[] = {
      { "a", "href" },
      { "img", "src" },
      { "img", "href" },
      { "body", "background" },
      { "base", "href" }, /* This one is special! */
      { NULL, NULL }
   };

   s = getstate(id);
   if (s->in_quote)
   {
      for (; *buf != '"' && bufsize > 0; buf++, bufsize--);
      s->in_quote = 0;
      if (*buf != '"')
      {
	 if (s->tag)
	    free(s->tag);
	 if (s->mod)
	    free(s->mod);
	 return NULL;
      }
      ++buf; --bufsize;
   }
   for (ptr = NULL; bufsize > 0; bufsize--, buf++)
   {
      switch (*buf)
      {
	 /* Since we're bound to be outside a quote,
	    '<' means entering a tag */
	 case '<':
	    s->in_tag = 1;
	    if (s->tag)
	    {
	       free(s->tag);
	       s->tag = NULL;
	    }
	    if (s->mod)
	    {
	       free(s->mod);
	       s->mod = NULL;
	    }
	    s->tg = s->md = 0;
	    break;
	 case '"':
	    if (s->in_quote)
	       s->in_quote = 0;
	    else if (s->in_tag)
	    {
	       s->in_quote = 1;
	       if (bufsize == 1)
	       {
		  if (s->tag)
		     free(s->tag);
		  if (s->mod)
		     free(s->mod);
		  return NULL;
	       }
	       ptr = ++buf;
	       --bufsize;
	       /* by testing *buf against newlines, I make sure the
		  parser doesn't crash at unclosed quotes. */
	       for (*size = 0;
		    *buf != '"' && *buf != '\n' && *buf != '#' && bufsize > 0;
		    ++*size, ++buf, --bufsize);
	       /* If the quoted word contains #, it is the HTML
		  marker, in which case *size is set to actually skip
		  the unimportant part.  if it is a bg color
		  specification, it will be skipped (*size == 0) */
	       if (bufsize && *size && *buf != '\n'
		   && idmatch(allow_tag, s->tag, s->mod))
	       {
		  if (!strcasecmp(s->tag, "base")
		      && !strcasecmp(s->mod, "href"))
		  {
		     if (s->base)
			free(s->base);
		     s->base = nmalloc(*size + 1);
		     strncpy(s->base, ptr, *size);
		     s->base[*size] = 0;
		  }
		  else
		     return ptr;
	       }
	       /* Now, if '#' had stopped the loop, get
		  on with it. */
	       if (*buf == '#' && !*size)
		  for (; *buf != '"' && *buf != '\n' && bufsize > 0;
		       ++buf, --bufsize);
	       if (bufsize == 1 || bufsize == 0)
	       {
		  if (s->tag)
		     free(s->tag);
		  if (s->mod)
		     free(s->mod);
		  return NULL;
	       }
	       if (*buf == '\n')
	       {
		  /* obviously no longer in quote. It might be well to
		     check whether '>' was encountered, but that would
		     be encouraging writers of invalid HTMLs, and we
		     don't want that, now do we? */
		  s->in_quote = 0;
		  break;
	       }
	       ++buf, --bufsize;
	       s->in_quote = 0;
	    }
	    break;
	 case ' ': case '\t': case '\n':
	    /* If a separator is found in tag, identifier is
	       finished. */
	    if (s->in_tag)
	       s->tg = s->md = 0;
	    break;
	 case '=': /* Ident separator + return processing */
	    if (s->in_tag)
	    {
	       s->md = s->tg = 0;
	       if (s->mod)
	       {
		  s->md = 0;
		  /* Added later:
		     links without quotes, like
		     <a href=http://fly.cc.fer.hr/>
		     1. skip spaces after '='
		     2. get the string (delimited by ' ', '\t',
		     '\n', '#' or '>') */
		  if (idmatch(allow_tag, s->tag, s->mod))
		  {
		     /* Separator skipping */
		     for (++buf, --bufsize;
			  bufsize > 1 && strchr(" \n\t", *buf);
			  ++buf, --bufsize)
			;
		     /* Return NULL if at EOF */
		     if (bufsize <= 1)
		     {
			if (s->tag)
			   free(s->tag);
			if (s->mod)
			   free(s->mod);
			return NULL;
		     }
		     /* If we've found a '"', give it up, and let the
		      appropriate switch deal with the bugger. */
		     if (*buf == '"')
		     {
			--buf, ++bufsize;
			break;
		     }
		     ptr = buf;
		     /* Hash support added, thanks to
			maverick@fly.cc.fer.hr */
		     for (*size = 0; bufsize > 1 && !strchr(" \n\t#>", *buf);
			  ++*size, ++buf, --bufsize)
			;
		     /* Return NULL if at EOF */
		     if (bufsize == 0)
		     {
			if (s->tag)
			   free(s->tag);
			if (s->mod)
			   free(s->mod);
			return NULL;
		     }
		     /* Again: take care of color specifications. */
		     if (!*size)
			break;
		     if (!strcasecmp(s->tag, "base")
			 && !strcasecmp(s->mod, "href"))
		     {
			if (s->base)
			   free(s->base);
			s->base = nmalloc(*size + 1);
			strncpy(s->base, ptr, *size);
			s->base[*size] = 0;
		     }
		     else
			/* That's it! */
			return ptr;
		  }
	       }
	    }
	    break;
	 case '>': /* Leave the tag */
	    if (s->tag)
	       free(s->tag);
	    s->tag = NULL;
	    if (s->mod)
	       free(s->mod);
	    s->mod = NULL;
	    s->in_tag = 0;
	    }
      if (s->in_tag && (isalnum(*buf) || *buf == '/' || *buf == '_'))
      {
	 if (!s->tag)
	    s->tg = 1;
	 else
	 {
	    /* If we already had a modifier, free it. */
	    if (s->mod && !s->md)
	    {
	       free(s->mod);
	       s->mod = NULL;
	    }
	    s->md = 1;
	 }
	 if (s->tg)
	    p = &(s->tag);
	 else
	    p = &(s->mod);
	 
	 if (!*p)
	 {
	    *p = nmalloc(2);
	    (*p)[0] = *buf;
	    (*p)[1] = 0;
	    s->idlen = 1;
	 }
	 else
	 {
	    ++(s->idlen);
	    *p = nrealloc(*p, s->idlen + 1);
	    (*p)[s->idlen - 1] = *buf;
	    (*p)[s->idlen] = 0;
	 }
      }
      }
   if (s->tag)
      free(s->tag);
   if (s->mod)
      free(s->mod);
   return NULL;
}

/* Get parser state, given a unique ID */
state_t *
getstate(char *id)
{
   static state_t *states = NULL;
   static int numstates = 0;
   int i;

   if (!numstates)
   {
      ++numstates;
      states = nmalloc(sizeof(state_t));
      init_state(states + numstates - 1, id);
      return states + numstates - 1;
   }
   else
   {
      for (i = 0; i < numstates; i++)
	 if (!strcmp(states[i].name, id))
	    break;
      if (i == numstates)
      {
	 ++numstates;
	 states = nrealloc(states, numstates * sizeof(state_t));
	 init_state(states + numstates - 1, id);
	 return states + numstates - 1;
      }
      else
	 return states + i;
   }
}

/* The function returns the base reference of HTML buffer id, or NULL
   if one wasn't defined for that buffer. */
char *
html_base(char *id)
{
   state_t *state;

   state = getstate(id);
   return state->base;
}

/* Initialize a parser state */
void
init_state(state_t *state, char *name)
{
   state->in_tag = state->in_quote = state->idlen = state->tg = state->md = 0;
   state->tag = state->mod = NULL;
   state->base = NULL;
   state->name = nstrdup(name);
}

==> geturl-1.3.1.tar.gz: geturl-1.3.1/html.c <==
/* html.c
   Module for parsing HTML buffers */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <ctype.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include "geturl.h"
#include "nmalloc.h"
#include "html.h"

/* Match a string against a null-terminated list of identifiers. */
int
idmatch(tag_t *tags, char *tag, char *mod)
{
   int i;

   if (!tag || !mod)
      return 0;
   
   for (i = 0; tags[i].tag; i++)
      if (!strcasecmp(tags[i].tag, tag) && !strcasecmp(tags[i].mod, mod))
	 return 1;
   return 0;
}

/* The following function parses an HTML buffer buf, of size
   bufsize searching for tags. When a tag is encountered, the
   routine will extract its components. If "src", "href" or
   "background" are found, their address is returned, and so
   is the length of the string. Returns NULL if no URL
   is found.

   You must also provide an id for the specified buffer, so
   that the function can associate it with a parser state.
   Thus several buffers can be parsed, and all the switching
   between them will be accurate, provided each of them has
   a unique ID (e.g. its filename).

   Latest: add no-quote support (e.g. href=http://something/)

   Note: possible bugs at the end of buffers, or with small
   buffers */

char *
htmlfindurl(char *buf, int bufsize, int *size, char *id)
{
   char *ptr, **p;
   state_t *s;
   /* NULL-terminated list of tags and modifiers someone would want to
      follow -- feel free to edit to suit your needs: */
   static tag_t allow_tag[] = {
      { "a", "href" },
      { "img", "src" },
      { "img", "href" },
      { "body", "background" },
      { "base", "href" }, /* This one is treated specially! */
      { NULL, NULL }
   };

   s = getstate(id);
   if (s->in_quote)
   {
      for (; *buf != '"' && bufsize > 0; buf++, bufsize--);
      s->in_quote = 0;
      if (*buf != '"')
      {
	 if (s->tag)
	    free(s->tag);
	 if (s->mod)
	    free(s->mod);
	 return NULL;
      }
      ++buf; --bufsize;
   }
   for (ptr = NULL; bufsize > 0; bufsize--, buf++)
   {
      switch (*buf)
      {
	 /* Since we're bound to be outside a quote,
	    '<' means entering a tag */
	 case '<':
	    s->in_tag = 1;
	    if (s->tag)
	    {
	       free(s->tag);
	       s->tag = NULL;
	    }
	    if (s->mod)
	    {
	       free(s->mod);
	       s->mod = NULL;
	    }
	    s->tg = s->md = 0;
	    break;
	 case '"':
	    if (s->in_quote)
	       s->in_quote = 0;
	    else if (s->in_tag)
	    {
	       s->in_quote = 1;
	       if (bufsize == 1)
	       {
		  if (s->tag)
		     free(s->tag);
		  if (s->mod)
		     free(s->mod);
		  return NULL;
	       }
	       ptr = ++buf;
	       --bufsize;
	       /* by testing *buf against newlines, I make sure the
		  parser doesn't crash at unclosed quotes. */
	       for (*size = 0;
		    *buf != '"' && *buf != '\n' && *buf != '#' && bufsize > 0;
		    ++*size, ++buf, --bufsize);
	       /* If the quoted word contains #, it is the HTML
		  marker, in which case *size is set to actually skip
		  the unimportant part.  if it is a bg color
		  specification, it will be skipped (*size == 0) */
	       if (bufsize && *size && *buf != '\n'
		   && idmatch(allow_tag, s->tag, s->mod))
	       {
		  if (!strcasecmp(s->tag, "base")
		      && !strcasecmp(s->mod, "href"))
		  {
		     if (s->base)
			free(s->base);
		     s->base = nmalloc(*size + 1);
		     strncpy(s->base, ptr, *size);
		     s->base[*size] = 0;
		  }
		  else
		     return ptr;
	       }
	       /* Now, if '#' had stopped the loop, get
		  on with it. */
	       if (*buf == '#' && !*size)
		  for (; *buf != '"' && *buf != '\n' && bufsize > 0;
		       ++buf, --bufsize);
	       if (bufsize == 1 || bufsize == 0)
	       {
		  if (s->tag)
		     free(s->tag);
		  if (s->mod)
		     free(s->mod);
		  return NULL;
	       }
	       if (*buf == '\n')
	       {
		  /* obviously no longer in quote. It might be well to
		     check whether '>' was encountered, but that would
		     be encouraging writers of invalid HTMLs, and we
		     don't want that, now do we? */
		  s->in_quote = 0;
		  break;
	       }
	       ++buf, --bufsize;
	       s->in_quote = 0;
	    }
	    break;
	 case ' ': case '\t': case '\n':
	    /* If a separator is found in tag, identifier is
	       finished. */
	    if (s->in_tag)
	       s->tg = s->md = 0;
	    break;
	 case '=': /* Ident separator + return processing */
	    if (s->in_tag)
	    {
	       s->md = s->tg = 0;
	       if (s->mod)
	       {
		  s->md = 0;
		  /* Added later:
		     links without quotes, like
		     <a href=http://fly.cc.fer.hr/>
		     1. skip spaces after '='
		     2. get the string (delimited by ' ', '\t',
		     '\n', '#' or '>') */
		  if (idmatch(allow_tag, s->tag, s->mod))
		  {
		     /* Separator skipping */
		     for (++buf, --bufsize;
			  bufsize > 1 && strchr(" \n\t", *buf);
			  ++buf, --bufsize)
			;
		     /* Return NULL if at EOF */
		     if (bufsize <= 1)
		     {
			if (s->tag)
			   free(s->tag);
			if (s->mod)
			   free(s->mod);
			return NULL;
		     }
		     /* If we've found a '"', give it up, and let the
		      appropriate switch deal with the bugger. */
		     if (*buf == '"')
		     {
			--buf, ++bufsize;
			break;
		     }
		     ptr = buf;
		     /* Hash support added, thanks to
			maverick@fly.cc.fer.hr */
		     for (*size = 0; bufsize > 1 && !strchr(" \n\t#>", *buf);
			  ++*size, ++buf, --bufsize)
			;
		     /* Return NULL if at EOF */
		     if (bufsize == 0)
		     {
			if (s->tag)
			   free(s->tag);
			if (s->mod)
			   free(s->mod);
			return NULL;
		     }
		     /* Again: take care of color specifications. */
		     if (!*size)
			break;
		     if (!strcasecmp(s->tag, "base")
			 && !strcasecmp(s->mod, "href"))
		     {
			if (s->base)
			   free(s->base);
			s->base = nmalloc(*size + 1);
			strncpy(s->base, ptr, *size);
			s->base[*size] = 0;
		     }
		     else
			/* That's it! */
			return ptr;
		  }
	       }
	    }
	    break;
	 case '>': /* Leave the tag */
	    if (s->tag)
	       free(s->tag);
	    s->tag = NULL;
	    if (s->mod)
	       free(s->mod);
	    s->mod = NULL;
	    s->in_tag = 0;
	    }
      if (s->in_tag && (isalnum(*buf) || *buf == '/' || *buf == '_'))
      {
	 if (!s->tag)
	    s->tg = 1;
	 else
	 {
	    /* If we already had a modifier, free it. */
	    if (s->mod && !s->md)
	    {
	       free(s->mod);
	       s->mod = NULL;
	    }
	    s->md = 1;
	 }
	 if (s->tg)
	    p = &(s->tag);
	 else
	    p = &(s->mod);
	 
	 if (!*p)
	 {
	    *p = nmalloc(2);
	    (*p)[0] = *buf;
	    (*p)[1] = 0;
	    s->idlen = 1;
	 }
	 else
	 {
	    ++(s->idlen);
	    *p = nrealloc(*p, s->idlen + 1);
	    (*p)[s->idlen - 1] = *buf;
	    (*p)[s->idlen] = 0;
	 }
      }
      }
   if (s->tag)
      free(s->tag);
   if (s->mod)
      free(s->mod);
   return NULL;
}

/* Get parser state, given a unique ID */
state_t *
getstate(char *id)
{
   static state_t *states = NULL;
   static int numstates = 0;
   int i;

   if (!numstates)
   {
      ++numstates;
      states = nmalloc(sizeof(state_t));
      init_state(states + numstates - 1, id);
      return states + numstates - 1;
   }
   else
   {
      for (i = 0; i < numstates; i++)
	 if (!strcmp(states[i].name, id))
	    break;
      if (i == numstates)
      {
	 ++numstates;
	 states = nrealloc(states, numstates * sizeof(state_t));
	 init_state(states + numstates - 1, id);
	 return states + numstates - 1;
      }
      else
	 return states + i;
   }
}

/* The function returns the base reference of HTML buffer id, or NULL
   if one wasn't defined for that buffer. */
char *
html_base(char *id)
{
   state_t *state;

   state = getstate(id);
   return state->base;
}

/* Initialize a parser state */
void
init_state(state_t *state, char *name)
{
   state->in_tag = state->in_quote = state->idlen = state->tg = state->md = 0;
   state->tag = state->mod = NULL;
   state->base = NULL;
   state->name = nstrdup(name);
}
